/*******************************************
ALU (Arithmetic and Logic Unit)
SR (Status Register)
UNIT
********************************************
HOMS version 1B
Hardware-Oriented Microprocessor
Simulator

(C) Panayotis (Panos) Papazoglou
https://homs.panospapazoglou.gr/

LICENSE: 
Creative Commons
CC BY NC SA
International License
********************************************/

//UNIT ADDRESS = 9
#define ADDRESS 9

//Include I2C library
#include <Wire.h>
#define FREQ 10000
//new tft section

#define ILI9341_BLACK 0x0000
#define ILI9341_BLUE 0x001F
#define ILI9341_WHITE 0xFFFF

#include "SPI.h"
#include "Adafruit_GFX.h"
#include "Adafruit_ILI9341.h"

//TFT pins
#define TFT_CLK 13
#define TFT_MISO 12
#define TFT_MOSI 11
#define TFT_DC 7
#define TFT_CS 10
#define TFT_RST 8

Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_MOSI, TFT_CLK, TFT_RST, TFT_MISO);

//This UNIT Receives/Sends commands from/to other units as bytes
#define READ 65
#define ADD 72
#define IDON 99
#define IDOFF 98
#define IDOK 97

#define GREEN_button 4
#define RED_button 5

//Variable that is used inside ISR
volatile byte rvTRUE = 0;

//Byte received from other units
byte rV;

//Store result of the numerical calculation
byte result = 0;

//Store the content of temporarily registers T1 and T2
byte T1 = 0;
byte T2 = 0;

//Flag array
byte SR[] = { 0, 0, 0 };  //VF (oVerflow), ZF (Zero), SF (Sign)

int bcolor = ILI9341_BLACK, fcolor = ILI9341_WHITE;


//Starting function (after reset or power on), runs once
void setup() {

  init_buttons();
  tft.begin();
  tft.setRotation(135);

  //Read POT for background color
  bcolor = analogRead(0);

  if (bcolor < 50) bcolor = ILI9341_BLACK;
  else if (bcolor > 975) bcolor = ILI9341_WHITE;
  else
    bcolor = ILI9341_BLUE;

  //Read POT for text color
  fcolor = analogRead(1);

  if (fcolor < 50) fcolor = ILI9341_BLACK;
  else if (fcolor > 975) fcolor = ILI9341_WHITE;
  else
    fcolor = ILI9341_YELLOW;

  tft.fillScreen(bcolor);

  //Start I2C communication
  Wire.begin(ADDRESS);
  Wire.setClock(FREQ);

  //Activate Receive & Request ISR routines
  Wire.onReceive(onReceive);
  Wire.onRequest(onRequest);

  //Display result and flag status
  DISPLAY_RES();
}

void loop() {
  //If a byte is received and is an ADD command,
  //then execute addition between the contents of Registers A and B
  //Store result in a variable, update SR[] array
  //and display result/SR contents

  if (rvTRUE == 1) {
    if (rV == ADD) {
      T1 = READ_REG(1);
      delay(100);
      T2 = READ_REG(2);
      result = T1 + T2;
      UPDATE_SR(result);
      DISPLAY_RES();
    }
    rvTRUE = 0;
  }
}

/*******************************************
ISR routine
Is activated when a byte is received
via I2C communication
********************************************/
void onReceive(int a) {
  rV = Wire.read();

  //Flag for activating code in order to read Reg. A and REg. B contents
  //I2C functions can not be called within an active I2C routine
  rvTRUE = 1;
}

/*******************************************
ISR routine
Is activated when a request is received
via I2C communication
********************************************/
void onRequest() {
  //Send result (ALU calculation) as answer to an I2C request

  if (rV == IDON) Wire.write(IDOK);
  else
    Wire.write(result);
}

/*******************************************
Read the content of a Register
with address addr
********************************************/
byte READ_REG(int addr) {
  byte ans;

  Wire.beginTransmission(addr);
  Wire.write(READ);
  Wire.endTransmission();

  //Receive answer
  Wire.requestFrom(addr, 1);
  if (Wire.available())
    ans = Wire.read();

  return ans;
}

/*******************************************
Update Status Register Flags based
on result status
********************************************/
void UPDATE_SR(byte res) {

  if (res == 0) SR[1] = 1;
  else SR[1] = 0;
  if (res < 0) SR[2] = 1;
  else SR[2] = 0;
  if (res > 255) {
    SR[0] = 1;
    result = 0;
  } else SR[0] = 0;
}

/*******************************************
Display result and Status Register Flags 
on LCD screen
********************************************/
void DISPLAY_RES() {

  display(60, 10, 5, 0, "ALU/SR");
  int y = 80;
  display(10, y, 4, 0, "RES10=");
  display(160, y, 4, 0, String(result));

  y += 50;
  display(10, y, 4, 0, "HEX=");
  display(110, y, 4, 0, String(result, HEX));

  y += 50;
  display(10, y, 4, 0, "V:");
  display(60, y, 4, 0, String(SR[0]));
  display(110, y, 4, 0, "Z:");
  display(160, y, 4, 0, String(SR[1]));
  display(210, y, 4, 0, "S:");
  display(260, y, 4, 0, String(SR[2]));
}

/*******************************************
Unit title
********************************************/
void tft_welcome() {
  tft.fillScreen(bcolor);
  display(30, 50, 4, ILI9341_GREEN, "Memory Unit");
}

/*******************************************
Display text on TFT 
on LCD screen

col: column
row: row
txsize: text size
txcolor: not used here (optional for debugging purposes)
val: text to be displayed
********************************************/
void display(int col, int row, int txsize, int txcolor, String val) {

  tft.setCursor(col, row);
  tft.setTextSize(txsize);
  tft.setTextColor(fcolor);
  tft.print(val);
}

/*******************************************
Initialize buttons by activating
internal pull-up resistors
********************************************/
void init_buttons() {
  pinMode(GREEN_button, INPUT_PULLUP);
  pinMode(RED_button, INPUT_PULLUP);
}
