/*******************************************
CONTROL UNIT
I2C MASTER
********************************************
HOMS version 1B
Hardware-Oriented Microprocessor
Simulator

(C) Panayotis (Panos) Papazoglou
https://homs.panospapazoglou.gr/

LICENSE: 
Creative Commons
CC BY NC SA
International License
********************************************/
//Please ignore any compilation warnings
//due to function overload (the same function name
//can be used but with different parameters).
//the MCU executes the function which matches
//with parameters

//Define symbolic names for sending commands to other blocks
//The I2C communication is based on one byte transmit/receive
#define READ 65
#define RESET 66
#define SHIFT_L 67
#define SHIFT_R 68
#define DEC 69
#define INC 70
#define LOAD 71
#define ADD 72
#define IDON 99
#define IDOFF 98
#define IDOK 97

//Set symbolic names for button/LED pins
#define GREEN_button 4
#define RED_button 5
//#define GREEN_LED 10
//#define RED_LED 9

//Include I2C library
//The I2C function are called using the prefix Wire.
#include <Wire.h>

#define FREQ 10000
//new tft section

#define ILI9341_BLACK 0x0000
#define ILI9341_RED 0xF800
#define ILI9341_WHITE 0xFFFF


#include "SPI.h"
#include "Adafruit_GFX.h"
#include "Adafruit_ILI9341.h"

//TFT pins
#define TFT_CLK 13
#define TFT_MISO 12
#define TFT_MOSI 11
#define TFT_DC 7
#define TFT_CS 10
#define TFT_RST 8

Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_MOSI, TFT_CLK, TFT_RST, TFT_MISO);

//Declare variables for the needs of the CU unit (locally)
byte RA = 0;
byte RB = 0;
byte RC = 0;
byte PC = 0;
byte MAR = 0;
byte MBR = 0;
byte T1 = 0;
byte T2 = 0;
byte RS = 0;
byte command = 0;
byte param = 0;
String hexval = "";
int blockADDR;

//Delay between execution process steps
int d = 3000;

int bcolor = ILI9341_BLACK, fcolor = ILI9341_WHITE;

//Starting function (after reset or power on), runs once
void setup() {
  //Initialize Serial communication
  //for displaying debugging messages
  Serial.begin(9600);

  init_buttons();
  tft.begin();
  tft.setRotation(135);

  //Read POT for background color
  bcolor = analogRead(0);
  Serial.print("bcolor=");
  Serial.println(bcolor);
  if (bcolor < 50) bcolor = ILI9341_BLACK;
  else if (bcolor > 975) bcolor = ILI9341_WHITE;
  else
    bcolor = ILI9341_RED;

  //Read POT for text color
  fcolor = analogRead(1);
  Serial.print("fcolor=");
  Serial.println(fcolor);
  if (fcolor < 50) fcolor = ILI9341_BLACK;
  else if (fcolor > 975) fcolor = ILI9341_WHITE;
  else
    fcolor = ILI9341_WHITE;

  //Initialize I2C communication
  Wire.begin();
  //Wire.setClock(FREQ);

  //Initialize buttons & LED-buttons
  init_buttons();

  Serial.println("init ok");

  //Display first message and read user input
  run();

  //Initial value for entering the while loop
  command = 0;

  //Clear registers (content=00)
  //Wait for RED button to start execution
  INIT();

  /****************************
  MAIN EXECUTION PROCESS
  *****************************/
  while (command != 17) {
    READ_PC();      //Read PC content for instruction starting address
    PC2MAR();       //Update MAR for accessibg memory address
    FETCH1();       //Fetch 1st byte from memory (instruction code)
    UPDATE_MAR();   //Update MAR for fetching 2nd byte
    FETCH2();       //Fetch 2nd byte (instruction parameter)
    DECODE_EXEC();  //Decode instruction and execute
    UPDATE_PC();    //Update PC content for next instruction

    next("Next >>");
  }

  //Out of while loop. The following code is executed when the instruction code is 17
  //which corresponds to STOP instruction
  tft.fillScreen(bcolor);
  display(50, 100, 3, 0, "END OF PROGRAM");
  display(50, 150, 3, 0, "EXECUTION");

}  //End of SETUP


/*******************************************
READ PC (Program Counter), PC block address=4
********************************************/

void READ_PC() {
  //Display message on LCD
  tft.fillScreen(bcolor);
  display(50, 30, 4, 0, "READING");
  display(40, 70, 3, 0, "Prog. Counter");
  display(40, 110, 3, 0, "(PC Register)");

  next("next>>");

  //Read PC content. Store content in PC variable
  PC = READ_REG(4);

  //Display READ results on LCD
  tft.fillScreen(bcolor);
  display(10, 10, 4, 0, "Prog.Counter");
  display(10, 60, 3, 0, "PC10=");
  display(100, 60, 3, 0, String(PC));
  display(10, 90, 3, 0, "PC16=");
  hexval = String(PC, HEX);
  display(100, 90, 3, 0, hexval);

  //Wait before next step
  next("next>>");

}  //End of READ_PC

/*******************************************
Store PC contents to register MAR
for accessing memory address
MAR=PC (MAR block address=5)
********************************************/
void PC2MAR() {
  //Display message on LCD
  tft.fillScreen(bcolor);
  display(20, 10, 4, 0, "MAR <-- PC");
  display(0, 60, 3, 0, "MAR=Memory Access Register");
  display(0, 125, 3, 0, "PC=Prog. Counter");
  next("next>>");

  //LOAD PC content in MAR register
  WRITE_REG(5, PC);

  //Update variable
  MAR = PC;

  //Display message on LCD
  tft.fillScreen(bcolor);
  display(10, 10, 4, 0, "Check MAR");
  display(0, 60, 3, 0, "MAR=Memory Access Register");
  next("next>>");
}  //End of PC2MAR

/*******************************************
FETCH 1st byte from memory, mem[MAR]
Instruction code
********************************************/
void FETCH1() {
  //Display message on LCD
  tft.fillScreen(bcolor);
  display(10, 10, 4, 0, "FETCH");
  display(140, 10, 3, 0, "byte 1>");
  display(10, 50, 3, 0, "Reading Instruction code");
  display(10, 110, 3, 0, "from mem[");
  display(175, 110, 3, 0, String(MAR));
  display(220, 110, 3, 0, "]");
  next("next>>");

  //LOAD MAR content in MEM/OUT Unit
  //for accessing the corresponding address
  WRITE_REG(11, MAR);

  //Receive memory content
  //Store content in MBR variable
  blockADDR = 11;
  Wire.requestFrom(blockADDR, 1);
  if (Wire.available())
    MBR = Wire.read();

  //Store instruction code in command variable
  command = MBR;

  //Display information (MAR, MBR) on LCD
  tft.fillScreen(bcolor);
  display(10, 10, 3, 0, "MBR<--mem[");
  display(190, 10, 3, 0, String(MAR));
  display(230, 10, 3, 0, "]");
  display(10, 40, 3, 0, "(10)=");
  display(100, 40, 3, 0, String(MBR));
  display(10, 70, 3, 0, "(16)=");
  hexval = String(MBR, HEX);
  display(100, 70, 3, 0, hexval);
  next("next>>");

  //Update MBR register from MBR variable
  WRITE_REG(6, MBR);

  //Display message on LCD
  tft.fillScreen(bcolor);
  display(10, 10, 4, 0, "Check MBR");
  next("next>>");
  tft.fillScreen(bcolor);
  display(10, 10, 4, 0, "command is");
  display(120, 50, 4, 0, String(command));
  next("next>>");
}  //End of FETCH1


/*******************************************
Updating MAR for accessing 
MAR=PC+1 (MAR address=5)
********************************************/
void UPDATE_MAR() {
  //Display message on LCD
  tft.fillScreen(bcolor);
  display(10, 10, 4, 0, "Updating");
  display(10, 60, 4, 0, "MAR...");

  //Update MAR register for fetching the next byte from memory
  //(from next address)
  WRITE_REG(5, PC + 1);
  next("next>>");

  //Update variable
  MAR = PC + 1;

  //Display message on LCD
  tft.fillScreen(bcolor);
  display(10, 10, 4, 0, "Check MAR");
  next("next>>");
}


/*******************************************
FETCH 2nd byte from memory, mem[MAR]
Parameter code
********************************************/
void FETCH2() {
  //Display message on LCD
  tft.fillScreen(bcolor);
  display(10, 10, 4, 0, "FETCH");
  display(140, 10, 3, 0, "byte 2>");

  //LOAD MAR content in MEM/OUT Unit
  //for accessing the corresponding address
  WRITE_REG(11, MAR);
  next("next>>");


  //Receive memory content
  //Store content in MBR variable
  blockADDR = 11;
  Wire.requestFrom(blockADDR, 1);
  if (Wire.available())
    MBR = Wire.read();

  //Store parameter code in param variable
  param = MBR;

  //Display information (MAR, MBR) on LCD
  tft.fillScreen(bcolor);
  display(10, 10, 3, 0, "MBR<--mem[");
  display(190, 10, 3, 0, String(MAR));
  display(230, 10, 3, 0, "]");
  display(10, 40, 3, 0, "(10)=");
  display(100, 40, 3, 0, String(MBR));
  display(10, 70, 3, 0, "(16)=");
  hexval = String(MBR, HEX);
  display(100, 70, 3, 0, hexval);
  next("next>>");

  //update register MBR
  WRITE_REG(6, MBR);

  //Display messages on LCD
  tft.fillScreen(bcolor);
  display(10, 10, 3, 0, "Check MBR");
  next("next>>");
  tft.fillScreen(bcolor);
  display(10, 10, 3, 0, "param is");
  display(180, 10, 3, 0, String(param));
  next("next>>");
}  //END of FETCH2


/*******************************************
Decoding instruction and execution
Instruction block: [command][param]

command=instruction code
param=parameter code
********************************************/
//command,param
void DECODE_EXEC() {

  /****************************
MOV A,i (A=i)
*****************************/
  if (command == 4) {
    //Display messages on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "CMD:MOV A,");
    display(190, 10, 3, 0, String(param));
    display(10, 40, 3, 0, "Executing...");
    next("next>>");

    //LOAD param (num) in Register A
    WRITE_REG(1, param);
  }


  /****************************
MOV B,i (B=i)
*****************************/
  if (command == 3) {
    //Display messages on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "CMD:MOV B,");
    display(100, 10, 3, 0, String(param));
    display(10, 40, 3, 0, "Executing...");
    next("next>>");

    //LOAD param (num) in Register A
    WRITE_REG(2, param);
  }


  /****************************
INC A (A=A+1)
*****************************/
  if (command == 10) {
    //Display messages on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "CMD:INC A");
    display(10, 40, 3, 0, "Executing...");
    next("next>>");

    //Send INC command to register A
    blockADDR = 1;
    Wire.beginTransmission(blockADDR);
    Wire.write(INC);
    Wire.endTransmission();

    //Display messages on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "Check REG-A");
  }

  /****************************
INC Β (Β=Β+1)
*****************************/
  if (command == 5) {
    //Display messages on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "CMD:INC Β");
    display(10, 40, 3, 0, "Executing...");
    next("next>>");

    //Send INC command to register Β
    blockADDR = 2;
    Wire.beginTransmission(blockADDR);
    Wire.write(INC);
    Wire.endTransmission();

    //Display messages on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "Check REG-Β");
  }

  /****************************
DEC A (A=A-1)
*****************************/
  if (command == 6) {
    //Display messages on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "CMD:DEC A");
    display(10, 40, 3, 0, "Executing...");
    next("next>>");

    //Send DEC command to register A
    blockADDR = 1;
    Wire.beginTransmission(blockADDR);
    Wire.write(DEC);
    Wire.endTransmission();

    //Display messages on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "Check REG-A");
  }


  /****************************
DEC B (B=B-1)
*****************************/
  if (command == 7) {
    //Display messages on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "CMD:DEC B");
    display(10, 40, 3, 0, "Executing...");
    next("next>>");

    //Send DEC command to register B
    blockADDR = 2;
    Wire.beginTransmission(blockADDR);
    Wire.write(DEC);
    Wire.endTransmission();

    //Display messages on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "Check REG-B");
  }

  /****************************
INC Β (Β=Β+1)
*****************************/
  if (command == 5) {
    //Display messages on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "CMD:INC Β");
    display(10, 40, 3, 0, "Executing...");
    next("next>>");

    //Send INC command to register Β
    blockADDR = 2;
    Wire.beginTransmission(blockADDR);
    Wire.write(INC);
    Wire.endTransmission();

    //Display messages on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "Check REG-Β");
  }


  /****************************
MOV B,A (B=A)
*****************************/
  if (command == 99)  //DISABLED INSTRUCTION
  {
    //Display messages on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "CMD:MOV B,A");
    display(10, 10, 3, 0, "Executing...");
    next("next>>");

    //Read from register A
    //Store content to variable RA
    RA = READ_REG(1);
    Serial.print("Read=RA:");
    Serial.println(RA);

    //Store register A content (RA) in register B
    WRITE_REG(2, RA);

    //Display message on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "Check REG-B");
  }

  /****************************
ADD A,B
*****************************/
  if (command == 1) {
    //Display message on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "CMD:ADD A,B");
    next("next>>");

    //READ from register A, Store in variable RA
    RA = READ_REG(1);

    //READ from register B, Store in variable RB
    RB = READ_REG(2);

    //Store register A content in register T1
    WRITE_REG(7, RA);

    //Store register B content in register T2
    WRITE_REG(8, RB);

    //Display messages on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "Check T1,T2");
    display(10, 40, 3, 0, "ALUEXEC...");
    next("next>>");

    //Send ADD instruction to ALU/SR Unit
    byte RES = SEND_ALU(9, ADD);

    //Store ALU result in register A
    WRITE_REG(1, RES);

    //Display message on LCD
    tft.fillScreen(bcolor);
    display(10, 10, 3, 0, "Check RA,ALU,SR");
    next("next>>");
  }


}  //End of DECODE_EXEC

/*******************************************
Update PC register content
for fetching next instruction bytes
from memory
********************************************/
void UPDATE_PC() {

  //Send two INC commands to PC register
  Wire.beginTransmission(4);
  Wire.write(INC);
  Wire.endTransmission();
  delay(50);
  Wire.beginTransmission(4);
  Wire.write(INC);
  Wire.endTransmission();
  delay(50);

  //Update variable
  PC = PC + 2;
}

void loop() {
  //Nothing here
}



/*******************************************
Initialize PULL-UP resistors
Every button is activated when
a PULL-UP pin goes to LOW (GND)
********************************************/
void init_buttons() {
  pinMode(GREEN_button, INPUT_PULLUP);
  pinMode(RED_button, INPUT_PULLUP);
}



/*******************************************
Clear all registers (set content=0)

Register address = 1 = Register A
Register address = 2 = Register B
Register address = 4 = Register PC
Register address = 5 = Register MAR
Register address = 6 = Register MBR
********************************************/
void clear_REGS() {
  Serial.println("Clearing registers...");
  int validADDR[] = { 1, 2, 4, 5, 6, 7 };
  for (int i = 0; i < 6; i++) {
    Wire.beginTransmission(validADDR[i]);
    Wire.write(RESET);
    Wire.endTransmission();
    delay(150);
  }
  Serial.println("OK done!");
}

/*******************************************
Display messages on TFT
and call clear register function
********************************************/
void INIT() {
  tft.fillScreen(bcolor);
  display(10, 10, 3, 0, "RESET REGISTERS");

  clear_REGS();
  display(10, 50, 3, 0, "Done!");
  display(10, 80, 3, 0, "Program loaded?");
  display(10, 110, 3, 0, "Start EXECUTION?");
  next("next>>");
}

/*******************************************
Read register content
Send READ command
Register address = addr
Return content
********************************************/
byte READ_REG(int addr) {
  byte ans;

  Wire.beginTransmission(addr);
  Wire.write(READ);
  Wire.endTransmission();
  delay(200);
  //Receive answer
  Wire.requestFrom(addr, 1);
  if (Wire.available())
    ans = Wire.read();

  return ans;
}

/*******************************************
Send command to ALU/SR unit
Address = addr
command to ALU/SR = instruction
Return ALU result
********************************************/
byte SEND_ALU(int addr, int instruction) {
  byte ans;

  Wire.beginTransmission(addr);
  Wire.write(instruction);
  Wire.endTransmission();
  delay(300);
  //Receive answer (PC content)

  Wire.requestFrom(addr, 1);
  if (Wire.available())
    ans = Wire.read();

  return ans;
}


/*******************************************
Load a number to a specific register
Register address = addr
Number = data
********************************************/
void WRITE_REG(byte addr, byte data) {

  Wire.beginTransmission(addr);
  Wire.write(LOAD);
  Wire.endTransmission();
  delay(200);
  Wire.beginTransmission(addr);
  Wire.write(data);
  Wire.endTransmission();
}


/*******************************************
Wait for a button press (LEFT button)
********************************************/
void next(String txt) {

  display(10, 180, 3, 0, txt);
  while (digitalRead(GREEN_button) == HIGH) { ; }
}

/*******************************************
Display first message and wait for user input
********************************************/
void run() {
  Serial.print("inside run");
  tft.fillScreen(bcolor);
  display(10, 20, 4, 0, "CONTROL UNIT");
  display(10, 70, 3, 0, "Initialize?");
  display(10, 100, 3, 0, "(Clear Registers)");


  next("next>>");
}

/*******************************************
Display text on TFT 
on LCD screen

col: column
row: row
txsize: text size
txcolor: not used here (optional for debugging purposes)
val: text to be displayed
********************************************/
void display(int col, int row, int txsize, int txcolor, String val) {

  tft.setCursor(col, row);
  tft.setTextSize(txsize);
  tft.setTextColor(fcolor);
  tft.print(val);
}
