/*******************************************
MEMORY & OUTPUT UNIT
********************************************
HOMS version 1B
Hardware-Oriented Microprocessor
Simulator

(C) Panayotis (Panos) Papazoglou
https://homs.panospapazoglou.gr/

LICENSE: 
Creative Commons
CC BY NC SA
International License
********************************************/

#define ILI9341_BLACK 0x0000
#define ILI9341_DARKGREEN 0x03E0
#define ILI9341_GREEN 0x07E0
#define ILI9341_WHITE 0xFFFF

#include "SPI.h"
#include "Adafruit_GFX.h"
#include "Adafruit_ILI9341.h"

//TFT pins
#define TFT_CLK 13
#define TFT_MISO 12
#define TFT_MOSI 11
#define TFT_DC 7
#define TFT_CS 10
#define TFT_RST 8

Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_MOSI, TFT_CLK, TFT_RST, TFT_MISO);

//Include I2C library
#include <Wire.h>
#define FREQ 10000

//Set symbolic names for button/LED pins
#define GREEN_button 4
#define RED_button 5


//A byte is read from memory when the LOAD command is received
#define LOAD 71
#define ADDRESS 11
#define IDON 99
#define IDOFF 98
#define IDOK 97


int bcolor = ILI9341_BLACK, fcolor = ILI9341_WHITE;

//Memory: 256 locations, starting address 0, ending address 255 (FF hex)
int mem[256];

//Serial received string
String rs = "";

//variables for serial communication
int i = 0;
int c = 0;
int r = 40;

//Received data from control unit
byte rV = 0;

//Previous received data from control unit
byte rVprev = 0;

//Flag that is activated when an address is received
byte ADDRflag = 0;

//Declare next array if you want to have a preloaded code in memory
byte prog[] = { 4, 3, 17, 0 }; //MOV A,4 ; HALT

//Variable for choosing demo mode or upload from PC
int demo = 0;


//Starting function (after reset or power on), runs once
void setup(void) {
  Serial.begin(9600);
  Serial.println("HELLO!");

  init_buttons();

  //Read POT for background color
  bcolor = analogRead(0);
  Serial.print("bcolor=");
  Serial.println(bcolor);
  if (bcolor < 50) bcolor = ILI9341_BLACK;
  else if (bcolor > 975) bcolor = ILI9341_WHITE;
  else
    bcolor = ILI9341_GREEN;

  //Read POT for text color
  fcolor = analogRead(1);
  Serial.print("fcolor=");
  Serial.println(fcolor);
  if (fcolor < 50) fcolor = ILI9341_BLACK;
  else if (fcolor > 975) fcolor = ILI9341_WHITE;
  else
    fcolor = ILI9341_BLACK;

  //initialize I2C communication (address 11)
  Wire.begin(ADDRESS);
  Wire.setClock(FREQ);

  //Initialize TFT
  tft.begin();
  tft.setRotation(135);

  tft_welcome();


  //Enable ISR routines for receiving bytes and requests
  Wire.onReceive(onReceive);
  Wire.onRequest(onRequest);

  //Initialize Serial communication
  //for displaying debugging messages
  Serial.begin(9600);
  init_mem(256);
  demo_or_upload();

  if (demo) {
    set_prog();

    tft.fillScreen(bcolor);
    display(50, 30, 4, 0, "D E M O");
    display(40, 70, 3, 0, "Instruction:");
    display(50, 100, 3, 0, "MOV A,3");
    display(20, 160, 3, 0, "Use CU to start");


    while (1);
  }

  tft.fillScreen(bcolor);
  display(50, 100, 3, ILI9341_WHITE, "Waiting for");
  display(70, 150, 3, ILI9341_WHITE, "Upload...");
}


/*******************************************
Main function (always active)
********************************************/
void loop(void) {

  //Section added for serial communication with the PC
  if (Serial.available() > 0) {
    rs = Serial.readStringUntil('\n');
    if (rs == "new") {
      tft.fillScreen(bcolor);

      i = 0;
      c = 0;
      display(15, 0, 3, ILI9341_BLUE, "Prog. bytes:");
      while (rs != "end") {
        if (Serial.available() > 0) {
          rs = Serial.readStringUntil('\n');
          mem[i] = rs.toInt();
          if (c > 300) {
            r += 40;
            c = 0;
          }

          if ((mem[i] == 0) && (i % 4 == 0)) {
            display(c, r, 2, ILI9341_WHITE, rs);
            c += 30;
          }
          if (mem[i] > 0) {
            display(c, r, 2, ILI9341_WHITE, rs);
            c += 30;
          }
          if (mem[i] == 17) rs = "end";
          i++;
        }
      }
      display(15, 100, 3, ILI9341_GREEN, "Upload complete");
      display(20, 150, 3, ILI9341_RED, "Use CU to start");
      rs = "stop";
    }

    else if (rs = "ready") {
      Serial.println("Sending from arduino...");
      for (int p = 0; p < i; p++) {
        //Serial.println("Hello from Arduino"); rs="stop";
        Serial.println(mem[p]);
      }
    }
  }
}





/*******************************************
initialize memory with size of locations

locations: total array locations
********************************************/
void init_mem(int locations) {
  for (int i = 0; i < locations; i++)
    mem[i] = 0;
}

/*******************************************
When data are received from control unit
********************************************/
void onReceive(int a)

{
  rV = Wire.read();
  Serial.print("rV:");
  Serial.println(rV);

  //If previous command is LOAD, then the current byte represents an address
  if (rVprev == LOAD) ADDRflag = 1;
  rVprev = rV;
}

/*******************************************
When data are requested from control unit
********************************************/
void onRequest() {
  //If the contents of an address are requested
  if (ADDRflag == 1) {
    Serial.println("ADDRflag=1");
    Serial.print("rV:");
    Serial.print(rV);
    Serial.print(", content:");
    Serial.println(mem[rV]);
    Wire.write(mem[rV]);
    ADDRflag = 0;
    rVprev = 0;
  }
  if (rV == IDON) Wire.write(IDOK);
}

/*******************************************
Preload program bytes in memory.
Update mem[] array from prog[] array
********************************************/
void set_prog() {
  for (int i = 0; i < 4; i++) mem[i] = prog[i];
}

/*******************************************
Unit title
********************************************/
void tft_welcome() {
  tft.fillScreen(bcolor);
  display(30, 50, 4, ILI9341_GREEN, "MEMORY UNIT");
}

/*******************************************
Display text on TFT 
on LCD screen

col: column
row: row
txsize: text size
txcolor: not used here (optional for debugging purposes)
val: text to be displayed
********************************************/
void display(int col, int row, int txsize, int txcolor, String val) {

  tft.setCursor(col, row);
  tft.setTextSize(txsize);
  tft.setTextColor(fcolor);
  tft.print(val);
}

/*******************************************
User selects Demo or Upload from PC
********************************************/
void demo_or_upload() {

  display(20, 100, 3, ILI9341_WHITE, "DEMO or UPLOAD?");
  display(50, 150, 2, ILI9341_YELLOW, "LEFT         RIGHT");
  display(20, 180, 2, ILI9341_GREEN, "(o) DEMO -- (o) UPLOAD?");

  int Gb = HIGH;
  int Rb = HIGH;
  while ((Gb == HIGH) && (Rb == HIGH)) {
    Gb = digitalRead(GREEN_button);
    Rb = digitalRead(RED_button);
  }
  if (Gb == LOW) demo = 1;
  else
    demo = 0;
}

/*******************************************
Wait for user input
********************************************/
void next(String txt) {
  display(0, 0, 4, ILI9341_WHITE, "(o) Next (o)--");
  while (digitalRead(GREEN_button) == HIGH) { ; }
}

/*******************************************
Initialize PULL-UP resistors
Every button is activated when
a PULL-UP pin goes to LOW (GND)
********************************************/
void init_buttons() {
  pinMode(GREEN_button, INPUT_PULLUP);
  pinMode(RED_button, INPUT_PULLUP);
}
